# Copyright (c) 2015-present, Facebook, Inc.
# All rights reserved.
"""
Train and eval functions used in main.py
"""
import math
import sys
from typing import Iterable, Optional

import torch

from timm.data import Mixup
from timm.utils import accuracy, ModelEma

from losses import DistillationLoss
import utils

import numpy as np
import os

def train_one_epoch(model: torch.nn.Module, criterion: DistillationLoss,
                    data_loader: Iterable, optimizer: torch.optim.Optimizer,
                    device: torch.device, epoch: int, loss_scaler, max_norm: float = 0,
                    model_ema: Optional[ModelEma] = None, mixup_fn: Optional[Mixup] = None,
                    set_training_mode=True, args = None):
    model.train(set_training_mode)
    metric_logger = utils.MetricLogger(delimiter="  ")
    metric_logger.add_meter('lr', utils.SmoothedValue(window_size=1, fmt='{value:.6f}'))
    header = 'Epoch: [{}]'.format(epoch)
    print_freq = 10


    model_without_ddp = model
    if args.distributed:
        model_without_ddp = model.module

    for samples, targets in metric_logger.log_every(data_loader, print_freq, header):
        samples = samples.to(device, non_blocking=True)
        targets = targets.to(device, non_blocking=True)

        model_without_ddp.target = targets
        model_without_ddp.save_images = False

        if mixup_fn is not None:
            samples, targets = mixup_fn(samples, targets)
            
        if args.bce_loss:
            targets = targets.gt(0.0).type(targets.dtype)
                    
        with torch.cuda.amp.autocast():
            outputs = model(samples)
            #loss = criterion(samples, outputs, targets)
            loss = criterion(samples, outputs[:,:,0], targets)
            for fix_idx in np.arange(1, model_without_ddp.num_fixations): #1, model.num_fixations):
                #loss = 0.9*loss + criterion(samples, outputs[:,:,fix_idx], targets)
                loss += criterion(samples, outputs[:,:,fix_idx], targets)

        loss_value = loss.item()

        if not math.isfinite(loss_value):
            print("Loss is {}, stopping training".format(loss_value))
            sys.exit(1)

        optimizer.zero_grad()

        # this attribute is added by timm on one optimizer (adahessian)
        is_second_order = hasattr(optimizer, 'is_second_order') and optimizer.is_second_order
        loss_scaler(loss, optimizer, clip_grad=max_norm,
                    parameters=model.parameters(), create_graph=is_second_order)

        torch.cuda.synchronize()
        if model_ema is not None:
            model_ema.update(model)

        metric_logger.update(loss=loss_value)
        metric_logger.update(lr=optimizer.param_groups[0]["lr"])
    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print("Averaged stats:", metric_logger)
    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}


@torch.no_grad()
def evaluate(data_loader, model, device, epoch, args_distributed):
    criterion = torch.nn.CrossEntropyLoss()

    metric_logger = utils.MetricLogger(delimiter="  ")
    header = 'Test:'

    # switch to evaluation mode
    model.eval()

    batch_idx = -1
    try:
        os.mkdir('saved_images/')
    except:
        pass

    model_without_ddp = model
    if args_distributed:
        model_without_ddp = model.module
    model_without_ddp.epoch = epoch
    try:
        os.mkdir('saved_images/'+str(epoch))
    except:
        pass

    model.time_taken = []

    STATS_DIR = 'Generated_results/Medium_sized_backbone/RG_Files/' #'Stats_Files/'
    try:
        os.mkdir(STATS_DIR)
    except:
        pass
    
    valid_subset = np.load('Generated_results/Medium_sized_backbone/RG_Files/valid_subet.npy')
    print(np.sum(valid_subset))

    for images, target in metric_logger.log_every(data_loader, 100, header):
        B = images.shape[0]
        batch_idx+=1

        if not valid_subset[batch_idx]:
            metric_logger.update(loss=0)#loss.item())
            continue
        images = images.to(device, non_blocking=True)
        target = target.to(device, non_blocking=True)

        model_without_ddp.batch_idx = batch_idx
        model_without_ddp.batch_cnt = batch_idx
        model_without_ddp.target = target
        model_without_ddp.save_images = False #True

        loss = 0
        #print('model: {}'.format(model))
        # compute output
        with torch.cuda.amp.autocast():
            output = model(images)
            #loss = criterion(output, target)
            loss = criterion(output[:,:,0], target)
            for fix_idx in np.arange(1, model_without_ddp.num_fixations):#1, model.num_fixations):
                #loss = 0.9*loss + criterion(output[:,:,fix_idx], target)
                loss += criterion(output[:,:,fix_idx], target)

        #acc1, acc5 = accuracy(output, target, topk=(1, 5))
        acc1_1, acc5_1 = accuracy(output[:,:,0], target, topk=(1, 5))
        acc1_2, acc5_2 = accuracy(output[:,:,1], target, topk=(1, 5))
        acc1_3, acc5_3 = accuracy(output[:,:,2], target, topk=(1, 5))
        acc1_4, acc5_4 = accuracy(output[:,:,3], target, topk=(1, 5))
        acc1_5, acc5_5 = accuracy(output[:,:,4], target, topk=(1, 5))

        batch_size = images.shape[0]
        metric_logger.update(loss=loss.item())
        metric_logger.meters['acc1_1'].update(acc1_1.item(), n=batch_size)
        metric_logger.meters['acc1_2'].update(acc1_2.item(), n=batch_size)
        metric_logger.meters['acc1_3'].update(acc1_3.item(), n=batch_size)
        metric_logger.meters['acc1_4'].update(acc1_4.item(), n=batch_size)
        metric_logger.meters['acc1_5'].update(acc1_5.item(), n=batch_size)
        #metric_logger.meters['acc5'].update(acc5.item(), n=batch_size)


    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print('* Fix@1 {top1.global_avg:.3f}'.format(top1=metric_logger.acc1_1))
    print('* Fix@2 {top1.global_avg:.3f}'.format(top1=metric_logger.acc1_2))
    print('* Fix@3 {top1.global_avg:.3f}'.format(top1=metric_logger.acc1_3))
    print('* Fix@4 {top1.global_avg:.3f}'.format(top1=metric_logger.acc1_4))
    print('* Fix@5 {top1.global_avg:.3f}'.format(top1=metric_logger.acc1_5))

    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}
